package org.dasein.cloud.cloudstack;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.security.SignatureException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.TreeSet;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.AbstractCloud;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.services.access.AccessServices;
import org.dasein.cloud.services.accounting.AccountingServices;
import org.dasein.cloud.services.address.AddressServices;
import org.dasein.cloud.services.cdn.ContentDistributionServices;
import org.dasein.cloud.services.dc.DataCenterServices;
import org.dasein.cloud.services.disk.SnapshotServices;
import org.dasein.cloud.services.disk.VolumeServices;
import org.dasein.cloud.services.firewall.FirewallServices;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.lb.LoadBalancerServices;
import org.dasein.cloud.services.scaling.ScalingServices;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.storage.StorageServices;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class CloudstackProvider extends AbstractCloud {
    
    public CloudstackProvider() { }
    
    String buildUrl(String command, Param ... params) throws SignatureException {
        try {
            StringBuilder str = new StringBuilder();
            String apiKey = new String(getContext().getAccessPublic(), "ISO-8859-1");
            String accessKey = new String(getContext().getAccessPrivate(), "ISO-8859-1");
            
            StringBuilder newKey = new StringBuilder();
            for( int i =0; i<apiKey.length(); i++ ) {
                char c = apiKey.charAt(i);
                
                if( c != '\r' ) {
                    newKey.append(c);
                }
            }
            apiKey = newKey.toString();
            newKey = new StringBuilder();
            for( int i =0; i<accessKey.length(); i++ ) {
                char c = accessKey.charAt(i);
                
                if( c != '\r' ) {
                    newKey.append(c);
                }
            }
            accessKey = newKey.toString();
            str.append(getHostUrl());
            str.append("/client/api?command=");
            str.append(command);
            for( Param param : params ) {
                str.append("&");
                str.append(URLEncoder.encode(param.getKey(), "ISO-8859-1"));
                str.append("=");
                str.append(URLEncoder.encode(param.getValue(), "ISO-8859-1"));
            }
            str.append("&apiKey=");
            str.append(URLEncoder.encode(apiKey, "ISO-8859-1"));
            str.append("&signature=");
            str.append(URLEncoder.encode(getSignature(command, apiKey, accessKey, params), "ISO-8859-1"));
            //System.out.println(str.toString());
            return str.toString();
        }
        catch( UnsupportedEncodingException e ) {
            e.printStackTrace();
            throw new RuntimeException("This cannot happen: " + e.getMessage());
        }
    }

    @Override
    public AccessServices getAccessServices() {
        return null;
    }

    @Override
    public AccountingServices getAccountingServices() {
        return null;
    }

    @Override
    public AddressServices getAddressServices() {
        return new IpAddress(this);
    }

    @Override
    public ContentDistributionServices getCdnServices() {
        return null;
    }

    @Override
    public String getCloudName() {
        String name = getContext().getCloudName();
        
        if( name == null ) {
            return "Cloud.com";
        }
        return name;
    }

    @Override
    public DataCenterServices getDataCenterServices() {
        return new Zones(this);
    }

    @Override
    public FirewallServices getFirewallServices() {
        return null;
    }

    private String getHostUrl() {
        ProviderContext ctx = getContext();
        
        return ctx.getEndpoint();
    }
    
    @Override
    public ImageServices getImageServices() {
        return new Templates(this);
    }

    @Override
    public LoadBalancerServices getLoadBalancerServices() {
        return new LoadBalancers(this);
    }

    @Override
    public String getProviderName() {
        String name = getContext().getProviderName();
        
        if( name == null ) {
            return "Cloud.com";
        }
        return name;
    }

    @Override
    public ScalingServices getScalingServices() {
        return null;
    }

    @Override
    public ServerServices getServerServices() {
        return new VirtualMachines(this);
    }

    private byte[] calculateHmac(String data, String key) throws SignatureException {
        try {
            SecretKeySpec signingKey = new SecretKeySpec(key.getBytes(), "HmacSHA1");
            Mac mac = Mac.getInstance("HmacSHA1");
            mac.init(signingKey);
        
            return mac.doFinal(data.getBytes());
        } 
        catch (Exception e) {
            throw new SignatureException("Failed to generate HMAC : " + e.getMessage());
        }
    }

    private String getSignature(String command, String apiKey, String accessKey, Param ... params) throws UnsupportedEncodingException, SignatureException {
        TreeSet<Param> sorted = new TreeSet<Param>();
        StringBuilder str = new StringBuilder();
        
        sorted.add(new Param("command", command.toLowerCase()));
        sorted.add(new Param("apikey", URLEncoder.encode(apiKey, "ISO-8859-1").toLowerCase()));
        for( Param param : params ) {
            sorted.add(new Param(URLEncoder.encode(param.getKey().toLowerCase(), "ISO-8859-1"), URLEncoder.encode(param.getValue().toLowerCase(), "ISO-8859-1")));
        }
        boolean first = true;
        for( Param param : sorted ) {
            if( !first ) {
                str.append("&");
            }
            first = false;
            str.append(param.getKey());
            str.append("=");
            str.append(param.getValue());
        }
        return new String(Base64.encodeBase64(calculateHmac(str.toString(), accessKey)));
    }
    
    @Override
    public SnapshotServices getSnapshotServices() {
        return new Snapshots(this);
    }

    @Override
    public StorageServices getStorageServices() {
        return null;
    }

    @Override
    public VolumeServices getVolumeServices() {
        return new Volumes(this);
    }
    
    Document parseResponse(InputStream responseBodyAsStream) throws CloudException, InternalException {
        return parseResponse(responseBodyAsStream, false);
    }
    
    Document parseResponse(InputStream responseBodyAsStream, boolean debug) throws CloudException, InternalException {
        try {
            BufferedReader in = new BufferedReader(new InputStreamReader(responseBodyAsStream));
            
            StringBuffer sb = new StringBuffer();
            String line;
            
            while ((line = in.readLine()) != null) {
                sb.append(line);
            }
            in.close();
            
            if( debug ) {
                System.out.println(sb.toString());
            }
            ByteArrayInputStream bas = new ByteArrayInputStream(sb.toString().getBytes());
            
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder parser = factory.newDocumentBuilder();
            Document doc = parser.parse(bas);
            bas.close();
            
            return doc;
        }
        catch( IOException e ) {
            throw new CloudException(e);
        }
        catch( ParserConfigurationException e ) {
            throw new CloudException(e);
        }
        catch( SAXException e ) {
            throw new CloudException(e);
        }
    }
    
    long parseTime(String timestamp) {
        DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ"); //2009-02-03T05:26:32.612278
        
        try {
            return df.parse(timestamp).getTime();
        }
        catch( ParseException e ) {
            df = new SimpleDateFormat("EEE MMM dd HH:mm:ss z yyyy"); //Sun Jul 04 02:18:02 EST 2010
            
            try {
                return df.parse(timestamp).getTime();
            }
            catch( ParseException another ) {
                return 0L;
            }
        }        
    }
    
    Document waitForJob(Document doc, String jobName) throws CloudException, InternalException {
        NodeList matches = doc.getElementsByTagName("jobid");
        if( matches.getLength() > 0 ) {
            return waitForJob(matches.item(0).getFirstChild().getNodeValue(), jobName);
        }    
        return null;
    }
    
    Document waitForJob(String jobId, String jobName) throws CloudException, InternalException {
        while( true ) {
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            HttpClient client = new HttpClient();
            GetMethod get = new GetMethod();
            Document doc;
            int code;
            
            try {
                get.setPath(buildUrl("queryAsyncJobResult", new Param[] { new Param("jobId", jobId) }));
                get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
            }
            catch( SignatureException e ) {
                throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
            }
            try {
                code = client.executeMethod(get);
            }
            catch( HttpException e ) {
                throw new InternalException("HttpException during GET: " + e.getMessage());
            }
            catch( IOException e ) {
                throw new CloudException("IOException during GET: " + e.getMessage());
            }
            if( code != HttpStatus.SC_OK ) {
                if( code == 401 ) {
                    throw new CloudException("Unauthorized user");
                }
                else if( code == 430 ) {
                    throw new InternalException("Malformed parameters");
                }
                else if( code == 431 ) {
                    throw new InternalException("Invalid parameters");
                }
                else if( code == 560 || code == 530 ) {
                    throw new CloudException("Server error in cloud (" + code + ")");
                }
                else if( code == 531 ) {
                    throw new CloudException("Unable to find account");
                }
                throw new CloudException("Received error code from server: " + code);
            }
            try {
                doc = parseResponse(get.getResponseBodyAsStream(), true);
            }
            catch( IOException e ) {
                throw new CloudException("IOException getting stream: " + e.getMessage());
            } 
            NodeList matches = doc.getElementsByTagName("jobstatus");
            int status = 0;
            
            if( matches.getLength() > 0 ) {
                status = Integer.parseInt(matches.item(0).getFirstChild().getNodeValue());
            }
            if( status > 0 ) {
                if( status == 1 ) {
                    return doc;
                }
                if( status == 2 ) {
                    matches = doc.getElementsByTagName("jobresult");
                    if( matches.getLength() > 0 ) {
                        throw new CloudException(matches.item(0).getFirstChild().getNodeValue());
                    }
                    else {
                        throw new CloudException(jobName + " failed with an unexplained error.");
                    }
                }
            }
        }
    }
}
